! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: elcplot_mod
!
! Authors: Natalie Webb - CESR, Ian Stewart - LUX
!          
! Description: Produces plots of the net source, background time series 
! and GTIs/flare set along with the fractional exposure of each point
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module elcplot_mod

  use caltypes	! for IntervalT
  use dal
  use param  
  use errorHandling
  use metatask_mod
  use dss
  use sla

  use temp_ssclib

  implicit none 
  
  type(TaskInfo) :: info ! type from metatask_mod

  type test
     real(kind=single) :: Stat, Prob
  end type test

  type data_tests
     real(kind=single) :: Mean_Count, Variance
     type(test) :: Chi_Square, K_S
  end type data_tests

  private 

  type localSrcData
     real(single)  :: rate, err
     logical       :: isNull
  end type localSrcData
  type localDataType
     real(double)       :: time
     real(single)       :: fracexp
     type(localSrcData) :: net, bkg
  end type localDataType

  public elcplot

  integer, parameter :: maxNumOutputBins = 512
  real, parameter :: timePadFrac = 0.03, yPadFrac3rdPlot = 0.1

contains 
  
  !---------------------------------------------------------------------
  subroutine elcplot
  !---------------------------------------------------------------------
  ! Input parameters :
  !   set : Source  series FITS file name - the output from lccorr
  !   plotdevice : Plot device type
  !   plotfile : Name of output file containing plots
  !   numInBinsPerOutBin : Number of bins to gather, default=5
  !   Units - the units that you want to plot the x-scale on (secs/days)
  !   bkgdyscale : Choice to plot background on same y-scale as counts
  ! Time variables :
  !   rawData%time : Vector containing middle of each time interval
  !   Bin_Width
  ! Counts variables :
  !   rawData%net%rate , ..%net%err , ..%bkg%rate , ..%bkg%err : Vectors containing
  !   respectively net count rate&error and background count rate&error.
  ! Label variable : Array containing the labels to add on plot

    character(len=128) :: Input_File, Output_File, Device_Type, GTI_File
    character(len=128) :: Flare_File
    integer :: numInBinsPerOutBin, numGtis, numOutputBins, numPoints, i, gtiNum
    character(len=4) :: Time_units
    real(single) :: Bin_Width
    character(30) :: DATE_OBS, meanNetRateStr, meanBkgRateStr
    type(localDataType), pointer :: rawData(:), plotData(:)
    type(IntervalT), pointer :: gtis(:)
    real(kind=single),dimension(:),pointer :: Flare_Rate
    real(double),dimension(:),pointer :: Flare_Time
    real ::&
      plotbinsize, netRateAtWindowBot, netRateAtWindowTop,&
      bkgRateAtWindowBot, bkgRateAtWindowTop, fracExpAtWindowBot,&
      fracExpAtWindowTop, addnAtWindowBot, addnAtWindowTop
    real(double) :: timeAtLHWindowEdge, timeAtRHWindowEdge, mjdOffsetDays,&
      offsetSec, mjdRefDays, offsetForPlottingPurposes,&
      trueTimeAtLHWindowEdge, trueTimeAtRHWindowEdge
    character(1000) :: Labels(9), Title
    real(kind=single) :: Window(3,4)
    character(5) :: boxXStr, boxYStr
    real, allocatable :: tVals(:), yVals(:)

    real           :: zzzWindow1(4),zzzWindow2(4),zzzWindow3(4),&
                      zzztimeAtLHWindowEdge,&
                      zzztimeAtRHWindowEdge
    character(20)  :: zzzboxXStr,&
                      zzzboxYStr,&
                      zzzxLabel,&
                      zzzyLabel




    Input_File = StringParameter("set")
    Device_Type = change_case(StringParameter("plotdevice"),2)
    Output_File = StringParameter("plotfile")
    numInBinsPerOutBin = IntParameter("binsize")
    Time_units = StringParameter("units")

    !...................................................................
    ! Read all the data:

    call read_file(Input_File, Bin_Width, mjdRefDays, DATE_OBS, rawData)

    if (booleanParameter("usegtiset") .and. booleanParameter("useflareset")) &
       call error('bothGtiAndFlare','Only GTI set or flare set ')
!**** a better way to arrange this would be to have parameters as follows:
! Enable withadditions
!   Choice additiontype
!     Page gti
!       Parameter gtiset
!     end
!     Page flare
!       Parameter flareset
!     end
!   end
! end

    meanNetRateStr = calcMeanRates(rawData%net%rate, rawData%net%isNull)
    meanBkgRateStr = calcMeanRates(rawData%bkg%rate, rawData%bkg%isNull)

    if (booleanParameter("usegtiset")) then
      GTI_File = StringParameter("gtiset")
      call readgtis(GTI_File, gtis)
      numGtis = size(gtis)
!**** should check that GTI file has same mjdref!!
    end if

    if (booleanParameter("useflareset")) then
      Flare_File = StringParameter("flareset")
      call readflare(Flare_File, Flare_Rate, Flare_Time)
!**** should check that flare file has same mjdref!!
    else
      allocate(Flare_Rate(0)) ! it gets passed to calcPgplotBoxWorldEdges
      ! even though it is not used therein unless --useflareset=yes.
    end if

    !...................................................................
    ! Calc times of plot bins:

    call rebin_ims(rawData, numInBinsPerOutBin, plotData)

    numOutputBins = size(plotData)
!    plotbinsize = (plotData(numOutputBins)%time - plotData(1)%time)&
!      / real(numOutputBins - 1)

    plotbinsize = Bin_Width*numInBinsPerOutBin
    !...................................................................
    ! Calc size of pgplot 'world' coordinates:

    call calcPgplotBoxWorldEdges(plotData, Flare_Rate, timeAtLHWindowEdge&
      , timeAtRHWindowEdge, netRateAtWindowBot, netRateAtWindowTop&
      , bkgRateAtWindowBot, bkgRateAtWindowTop, fracExpAtWindowBot&
      , fracExpAtWindowTop, addnAtWindowBot, addnAtWindowTop)

    bkgRateAtWindowBot = 0.0 - epsilon(1.0) ! just so that rounding errors
    ! don't prevent the 0 being shown at the base of the Y scale.

    !...................................................................
    ! Insert dummy rate values at nulls (so these will be plotted well
    ! out of window):

    where(plotData%net%isNull)
      plotData%net%err  = 0.0
      plotData%net%rate = netRateAtWindowTop&
        + 100.0 * abs(netRateAtWindowTop - netRateAtWindowBot)
    end where
    where(plotData%bkg%isNull)
      plotData%bkg%err  = 0.0
      plotData%bkg%rate = bkgRateAtWindowTop&
        + 100.0 * abs(bkgRateAtWindowTop - bkgRateAtWindowBot)
    end where
    ! The '100' is arbitrary.

    !...................................................................
    ! Calculate offset and apply it to all time data:

    call calcTimeOffset(mjdRefDays, plotData(1)%time, DATE_OBS&
      , mjdOffsetDays, offsetSec)

    ! Originally I subtracted the value of offsetSec from all time
    ! values. However this turns out to be not so good if the resulting
    ! times are too large (either large positive or large negative).
    ! This is because pgplot only accepts single-precision arguments,
    ! thus submitting eg time values which span a small range compared
    ! to their average can stuff up both the plot and the printed X
    ! scale. Best is to always use relative time values (eg offset by
    ! lowest time, though highest time or indeed average time would do
    ! as well) for both the x values of any plotted points and the X
    ! boundaries of the pgplot window (pgswin call). However this strategy
    ! if fully pursued would necessitate the messy printing 'by hand'
    ! of the time label axes, instead of relying on pgbox. Haven't got
    ! time to write such a routine at the moment so I am doing a
    ! 2-bites-at-the-cherry approach within make_one_plot plus have
    ! added some warnings when the time values get out of range.

    trueTimeAtLHWindowEdge = timeAtLHWindowEdge - offsetSec
    trueTimeAtRHWindowEdge = timeAtRHWindowEdge - offsetSec

    offsetForPlottingPurposes = plotData(1)%time
    plotData%time = plotData%time - offsetForPlottingPurposes
    if (booleanParameter("usegtiset")) then
      gtis%lower = gtis%lower - offsetForPlottingPurposes
      gtis%upper = gtis%upper - offsetForPlottingPurposes
    end if
    if (booleanParameter("useflareset")) then
      Flare_Time = Flare_Time - offsetForPlottingPurposes
    end if
    timeAtLHWindowEdge = timeAtLHWindowEdge - offsetForPlottingPurposes
    timeAtRHWindowEdge = timeAtRHWindowEdge - offsetForPlottingPurposes

    !...................................................................
    ! Apply time-axis units to all time data:

    if (Time_units == 'days') then
      plotData%time = plotData%time / 24.0d0 / 3600.0d0
      if (booleanParameter("usegtiset")) then
        gtis%lower = gtis%lower / 24.0d0 / 3600.0d0
        gtis%upper = gtis%upper / 24.0d0 / 3600.0d0
      end if
      if (booleanParameter("useflareset")) then
        Flare_Time = Flare_Time / 24.0d0 / 3600.0d0
      end if
      timeAtLHWindowEdge = timeAtLHWindowEdge / 24.0d0 / 3600.0d0
      timeAtRHWindowEdge = timeAtRHWindowEdge / 24.0d0 / 3600.0d0
      trueTimeAtLHWindowEdge = trueTimeAtLHWindowEdge / 24.0d0 / 3600.0d0
      trueTimeAtRHWindowEdge = trueTimeAtRHWindowEdge / 24.0d0 / 3600.0d0
    end if

    if (trueTimeAtLHWindowEdge <= -2.0e8) call warning('badLabelsLowerTime'&
      ,"The time (minus offset) at the left-hand edge of the plot is smaller &
      &than -2.0e8 in the chosen units. Unfortunately the plotting subroutine &
      &does not handle well values in this range. For the time being (until &
      &we fix this behaviour), if you want sensible numbers printed, you &
      &will have to use a different offset value, such that the start time &
      &becomes greater than -2.0e8")
    if (trueTimeAtRHWindowEdge >= 2.0e8) call warning('badLabelsUpperTime'&
      ,"The time (minus offset) at the right-hand edge of the plot is larger &
      &than 2.0e8 in the chosen units. Unfortunately the plotting subroutine &
      &does not handle well values in this range. For the time being (until &
      &we fix this behaviour), if you want sensible numbers printed, you &
      &will have to use a different offset value, such that the end time &
      &becomes smaller than 2.0e8")

    !...................................................................
    ! Do the plot:

    call create_labels(Input_File, plotbinsize, mjdOffsetDays, Labels)
    call open_plot(Device_Type , Output_File)
    call plotHeaderInfo(Labels)

    Window(1,:) = (/0.1,0.94,0.66,0.84/)
    Window(2,:) = (/0.1,0.94,0.38,0.56/)
    Window(3,:) = (/0.1,0.94,0.1,0.28/)

    Title = 'Background subtracted time series'//'            Mean Rate='&
      // trim(meanNetRateStr)
zzzWindow1 = Window(1,1:4)
zzzWindow2 = Window(2,1:4)
zzzWindow3 = Window(3,1:4)
zzztimeAtLHWindowEdge = real(timeAtLHWindowEdge)
zzztimeAtRHWindowEdge = real(timeAtRHWindowEdge)
zzzboxXStr = 'BCNST'
zzzboxYStr = 'BCNST'
zzzxLabel = 'Time ('//Time_units//') '
zzzyLabel = 'Counts/sec'

    call make_one_plot(zzzWindow1&
      , real(timeAtLHWindowEdge), real(timeAtRHWindowEdge)&
      , trueTimeAtLHWindowEdge, trueTimeAtRHWindowEdge&
      , netRateAtWindowBot, netRateAtWindowTop&
      , 'BCNST', 'BCNST', xLabel='Time ('//Time_units//') '&
      , yLabel='Counts/sec', Title=Title)

    call pgsci(2)
    call pgerry(numOutputBins, real(plotData%time)&
      , real(plotData%net%rate + plotData%net%err)&
      , real(plotData%net%rate - plotData%net%err), 0.0)
    call pgsci(1)
    call pgslw(5)
    call pgpt(numOutputBins, real(plotData%time), real(plotData%net%rate), -1)


    Title = 'Background time series'//'                        Mean Rate='&
      // trim(meanBkgRateStr)
    call make_one_plot(zzzWindow2&
      , zzztimeAtLHWindowEdge, zzztimeAtRHWindowEdge&
      , trueTimeAtLHWindowEdge, trueTimeAtRHWindowEdge&
      , bkgRateAtWindowBot, bkgRateAtWindowTop&
      , zzzboxXStr, zzzboxYStr, xLabel=zzzxLabel&
      , yLabel='Counts/sec', Title=Title)

    call pgsci(2)
    call pgerry(numOutputBins, real(plotData%time)&
      , real(plotData%bkg%rate + plotData%bkg%err)&
      , real(plotData%bkg%rate - plotData%bkg%err), 0.0)
    call pgsci(1)
    call pgslw(5)
    call pgpt(numOutputBins, real(plotData%time), real(plotData%bkg%rate), -1)


    if (booleanParameter("usegtiset")) then
      Title = 'GTI time series (histogram) and fractional exposure &
        &values (blue points)'
      boxXStr = ''
      boxYStr = 'CMST'
    else if (booleanParameter("useflareset")) then
      Title = 'Background flare time series and fractional exposure &
        &values (blue points)'
      boxXStr = ''
      boxYStr = 'CMST'
    else
      Title = 'Fractional Exposure'
      boxXStr = 'BCNST'
      boxYStr = 'BCMST'
    end if

    call make_one_plot(zzzWindow3&
      , zzztimeAtLHWindowEdge, zzztimeAtRHWindowEdge&
      , trueTimeAtLHWindowEdge, trueTimeAtRHWindowEdge&
      , fracExpAtWindowBot, fracExpAtWindowTop&
      , boxXStr, boxYStr, xLabel=zzzxLabel&
      , yLabel='', Title=Title)

    call pgsci(4)
    call pgmtxt('R', 2.5, 0.4, 0.0, 'FRACEXP') 
    call pgslw(5)
    call pgpt(numOutputBins, real(plotData%time), real(plotData%fracexp), -1)


    if (booleanParameter("usegtiset")) then
      call make_one_plot(Window(3,:)&
      , real(timeAtLHWindowEdge), real(timeAtRHWindowEdge)&
      , trueTimeAtLHWindowEdge, trueTimeAtRHWindowEdge&
        , addnAtWindowBot, addnAtWindowTop&
        , 'BCNST', 'BNT', xLabel='', yLabel='GTI', Title='')

      call pgslw(1)
      call pgsci(1)

      ! Construct the sequence of line segments:

      numPoints = 2 + 4 * numGtis
      allocate(tVals(numPoints), yVals(numPoints))
      tVals(1)         = real(timeAtLHWindowEdge&
        - 100.0d0 * (timeAtRHWindowEdge-timeAtLHWindowEdge))
      tVals(numPoints) = real(timeAtRHWindowEdge&
        + 100.0d0 * (timeAtRHWindowEdge-timeAtLHWindowEdge))
      ! The '100' is arbitrary.

      yVals(1)         = 0.0
      yVals(numPoints) = 0.0

      do gtiNum = 1, numGtis
        i = 1 + 4 * (gtiNum - 1)
        tVals(i + 1) = real(gtis(gtiNum)%lower)
        tVals(i + 2) = real(gtis(gtiNum)%lower)
        tVals(i + 3) = real(gtis(gtiNum)%upper)
        tVals(i + 4) = real(gtis(gtiNum)%upper)

        yVals(i + 1) = 0.0
        yVals(i + 2) = 1.0
        yVals(i + 3) = 1.0
        yVals(i + 4) = 0.0
      end do

      call pgline(numPoints, tVals, yVals)

    else if (booleanParameter("useflareset")) then
      call make_one_plot(Window(3,:)&
        , real(timeAtLHWindowEdge), real(timeAtRHWindowEdge)&
        , trueTimeAtLHWindowEdge, trueTimeAtRHWindowEdge&
        , addnAtWindowBot, addnAtWindowTop&
        , 'BCNST', 'BNST', xLabel='', yLabel='Flare count rate', Title='')

      call pgslw(1)
      call pgsci(1)
      call pgline(size(Flare_Time), real(Flare_Time), real(Flare_Rate))
    end if

    call pgclos


    deallocate(rawData, plotData)
    if (booleanParameter("usegtiset")) then
      deallocate(gtis)
    end if
    if (booleanParameter("useflareset")) then
      deallocate(Flare_Rate, Flare_Time)
    end if

  end subroutine elcplot

  !-------------------------------------------------------------------------------------
  subroutine read_file(Input_File, Bin_Width, mjdRefDays, DATE_OBS, rawData)
  !-------------------------------------------------------------------------------------
  ! Reads source time series FITS file and recovers all data needed. Also
  ! creates time vector.

    character(*),        intent(in)  :: Input_File
    real(single),        intent(out) :: Bin_Width
    real(double),        intent(out) :: mjdRefDays
    character(30),       intent(out) :: DATE_OBS
    type(localDataType), pointer     :: rawData(:)

  ! Local variables are Set, Table, Column and Pointer needed by DAL to 
  ! recover FITS binary table.Time_Zero contains the start exposure time in MJD

    type(DataSetT) :: Ts_Set
    type(TableT) :: Ts_Table
    type(ColumnT) :: Col_Rate, Col_Error, Col_Backv, Col_Backe
    type(ColumnT) :: Col_Frac, Col_Time
    real(kind=single),dimension(:),pointer :: Array_Rate, Array_Error
    real(kind=single),dimension(:),pointer :: Array_Backv, Array_Backe
    real(kind=single),dimension(:),pointer ::  Array_Frac
    real(kind=double),dimension(:),pointer ::  Array_Time
    real(kind=single) :: Time_Zero
    integer :: Nbin
    integer :: i
    real(double) :: centreTime1stBin
    character(len=12) :: backapp


    call read_data (Input_File, Ts_Set, backapp)!, Content
    if (trim(backapp) == 'F')&
      call warning('backgroundNotSubtracted', 'Non background &
        &subtracted lightcurve supplied')

! The keywords TSTART, TIMEDEL, and EXPOSURE TIME are in the RATE extension 
! not anymore in the primary header

    call read_table(Input_File, Ts_Set, Ts_Table, Time_Zero, Bin_Width)
!, Exposure_Time)

    Col_Time  = Column(Ts_Table, ColumnNumber(Ts_Table,'TIME'),Read)
    Col_Rate  = Column(Ts_Table, ColumnNumber(Ts_Table,'RATE'),Read)
    Col_Error = Column(Ts_Table, ColumnNumber(Ts_Table,'ERROR'),Read)
    Col_Backv = Column(Ts_Table, ColumnNumber(Ts_Table,'BACKV'),Read)
    Col_Backe = Column(Ts_Table, ColumnNumber(Ts_Table,'BACKE'),Read)
    Col_Frac  = Column(Ts_Table, ColumnNumber(Ts_Table,'FRACEXP'),Read)

    Array_Time  => real64Data(Col_Time)
    Array_Rate  => real32Data(Col_Rate)
    Array_Error => real32Data(Col_Error)
    Array_Backv => real32Data(Col_Backv)
    Array_Backe => real32Data(Col_Backe)
    Array_Frac  => real32Data(Col_Frac)

    Nbin = numberOfRows(Ts_Table)
    if (Nbin <= 0)&
call error('noDataInTimeSeries',"The RATE table of the time series &
&contains no rows")

    allocate(rawData(Nbin))

    rawData%time = Array_Time
    rawData%net%rate  = Array_Rate
    rawData%net%err   = Array_Error
    rawData%bkg%rate  = Array_Backv
    rawData%bkg%err   = Array_Backe
    rawData%fracexp  = Array_Frac

    rawData%net%isNull = .false. ! default
    rawData%bkg%isNull = .false. ! default
    do i=1,Nbin
      if (isNull(Col_Rate,  i-1)) rawData(i)%net%isNull = .true.
      if (isNull(Col_Backv, i-1)) rawData(i)%bkg%isNull = .true.
    end do

    where(isNaN(rawData%net%rate)) rawData%net%isNull = .true.
    where(isNaN(rawData%bkg%rate)) rawData%bkg%isNull = .true.
    ! see module temp_ssclib for isNaN

    if (all(rawData%net%isNull))&
call error('allRateValuesAreNull', "All of the RATE values in the &
&time series seem to be nulls or NaNs.")

    ! If the non-background subtracted lightcurve is provided, subtract the bkg

    if (trim(backapp) == 'F') then 
      if (minval(rawData%net%rate, .not.rawData%net%isNull)  < 0.0 ) &
call error('negativeRawSourceCounts','Supposedly the background has NOT &
&been subtracted from the source counts... but negative source &
&counts have been detected!')

      do i=1,Nbin
        if (rawData(i)%bkg%isNull) then
          rawData(i)%net%isNull = .true.
        else
          rawData(i)%net%rate = rawData(i)%net%rate - rawData(i)%bkg%rate
          rawData(i)%net%err = sqrt(rawData(i)%net%err*rawData(i)%net%err&
            + rawData(i)%bkg%err*rawData(i)%bkg%err)
        end if
      end do

      if (all(rawData%net%isNull))&
call error('allTsValuesAreNull', "There doesn't seem to be any bin &
&in the time series for which both RATE and BACKV are both non-null.")
    end if

    if (all(rawData%bkg%isNull)) then
      call warning('allBackvValuesAreNull',"All of the BACKV values in the &
        &time series seem to be nulls or NaNs.")
    else
      if (minval(rawData%bkg%rate, .not.rawData%bkg%isNull)  < 0.0 ) &
call error('negativeBackgroundRate','Background rate negative')
      if (minval(rawData%bkg%err,  .not.rawData%bkg%isNull)  < 0.0 ) &
call error('negativeBackgroundError','Background error rate negative')
    end if

    if ( minval(rawData%net%err, .not.rawData%net%isNull) < 0.0) &
call error('negativeSourceError','Count rate errors negative')

    centreTime1stBin = calcCentreTime1stBin(Ts_Table)
    rawData%time = rawData%time + centreTime1stBin

    call release(Col_Time)
    call release(Col_Rate)
    call release(Col_Error)
    call release(Col_Backv)
    call release(Col_Backe)
    call release(Col_Frac) 

    mjdRefDays = real64Attribute(Ts_Table, 'MJDREF')

    if (HasAttribute(Ts_Set,"DATE-OBS") ) then
      DATE_OBS = StringAttribute(Ts_Set,'DATE-OBS')
    else if (HasAttribute(Ts_Table,"DATE-OBS") ) then
      DATE_OBS = StringAttribute(Ts_Table,'DATE-OBS')
    else
      DATE_OBS = 'Indef'
    endif

    call release(Ts_Table)    
    call release(Ts_Set)

  end subroutine read_file

  !-------------------------------------------------------------------------------------
  subroutine read_data (Input_File, Ts_Set, BACKAPP)!, Content
  !-------------------------------------------------------------------------------------
  ! Reads the in parameter FITS file and gives back a handle to 
  ! corresponding dataset. Checks some keywords existence and validity.

    character(len=*),intent(in) :: Input_File 
    type( DataSetT ),intent(out) :: Ts_Set
!    character(len=*),intent(out) :: Content
    character(len=12),intent(out) :: BACKAPP

    type( TableT ) :: Table1

    Ts_Set = dataSet(Input_File,READ)
    Table1 = table(Ts_Set,BlockNumber(Ts_Set,'RATE')) 

    if (HasAttribute(Ts_Set,"BACKAPP") ) then
         backapp = StringAttribute(Ts_Set,'BACKAPP')
    else if (HasAttribute(Table1,"BACKAPP") ) then
         backapp = StringAttribute(Table1,'BACKAPP')
    else
         backapp = 'F'
    endif

    if ( .not. HasAttribute(Ts_Set,"CONTENT") ) then
      call warning('noContentKeyword','Keyword CONTENT missing '//Input_File)
!        Content = 'SOURCE TIMESERIES'
!    else
!        Content = change_case(StringAttribute(Ts_Set,"CONTENT"),2)
    endif

    call release(Table1)

  end subroutine read_data

  !-------------------------------------------------------------------------------------
  subroutine read_table (Input_File, Ts_Set, Ts_Table, Time_Zero, Bin_Width)
!, Exposure_Time )
  !-------------------------------------------------------------------------------------
  ! With FITS file name and handle to dataset, generates table.
  ! Checks more keywords.
  ! Also gives back Time_Zero, Bin_Width, Exposure_Time and Content.

    character(len=*),intent(in) :: Input_File
    type( DataSetT ),intent(in) :: Ts_Set
    type( TableT ),intent(out) :: Ts_Table

    real(kind=single),intent(out) :: Time_Zero, Bin_Width
! , Exposure_Time

    Ts_Table = table(Ts_Set,BlockNumber(Ts_Set,'RATE'))
    if ( .not. HasAttribute(Ts_Table,"HDUCLASS") ) &
      call warning('noKeyword','Keyword HDUCLASS missing '//Input_File)
    if ( change_case(StringAttribute(Ts_Table,"HDUCLASS"),2) /= 'OGIP' ) &
      call warning('badKeyword','HDUCLAS is not equal to OGIP')
    if ( .not. HasAttribute(Ts_Table,"HDUCLAS1") ) &
      call warning('noKeyword','Keyword HDUCLAS1 missing  '//Input_File)
    if ( change_case(StringAttribute(Ts_Table,"HDUCLAS1"),2) /= 'LIGHTCURVE' ) &
      call warning('badKeyword','HDUCLAS1 is not equal to LIGHTCURVE')
    if ( .not. HasAttribute(Ts_Table,"HDUCLAS2") ) &
      call warning('noKeyword','Keyword HDUCLAS2 missing '//Input_File)
!    if ( change_case(StringAttribute(Ts_Table,"HDUCLAS2"),2) /= 'TOTAL' ) &
!      call warning('ELCPLOT','HDUCLAS2 is not equal to TOTAL')
    if ( .not. HasAttribute(Ts_Table,"HDUCLAS3")) &
      call warning('noKeyword','Keyword HDUCLAS3 missing '//Input_File)
    if ( change_case(StringAttribute(Ts_Table,"HDUCLAS3"),2) /= 'RATE' ) &
      call warning('badKeyword','HDUCLAS3 is not equal to RATE')

    if ( .not. HasAttribute(Ts_Table,"TSTART") ) then
      call warning('noKeyword','Keyword TSTART missing '//Input_File)
    else
      Time_Zero = real32Attribute(Ts_Table,'TSTART')
    endif

    if ( .not. HasAttribute(Ts_Table,"TIMEDEL") ) &
      call error('noTimedelKeyword','Keyword TIMEDEL missing '//Input_File)
    Bin_Width = real32Attribute(Ts_Table,'TIMEDEL')
    if ( Bin_Width < epsilon(0.) ) call error('badBinWidth', &
      'Null or negative Bin_Width '//Input_File)

!    if ( .not. HasAttribute(Ts_Table,"EXPOSURE") ) then
!         call warning('ELCPLOT','Keyword EXPOSURE missing '//Input_File)
!    else
!         Exposure_Time = real32Attribute(Ts_Table,"EXPOSURE") 
!         if ( Exposure_Time < epsilon(0.) ) call warning('ELCPLOT', &
!          'Exposure time < 0 '//Input_File)
!    endif

  end subroutine read_table

  !---------------------------------------------------------------------
  function calcCentreTime1stBin(tab) result(centreTime1stBin)
  !---------------------------------------------------------------------
  ! Taken from lccorr

    implicit none 

    type(TableT), intent(in) :: tab
    real(double) :: centreTime1stBin

    real(double) ::&
      tsTimeZerF
    integer(int32) ::&
      tsTimeZerI

    !-------------------------------------------------------------------
    ! Start:

    if (hasAttribute(tab, 'TIMEZERI')&
    .and. hasAttribute(tab, 'TIMEZERF')) then
      tsTimeZerI = int32Attribute(tab, 'TIMEZERI')
      tsTimeZerF = real64Attribute(tab, 'TIMEZERF')
      centreTime1stBin = tsTimeZerI + tsTimeZerF ! type conversion to real64

    else if (hasAttribute(tab, 'TIMEZERO')) then
      centreTime1stBin = real64Attribute(tab, 'TIMEZERO')

    else
      centreTime1stBin = 0.0d0
    end if

  end function calcCentreTime1stBin

  !---------------------------------------------------------------------
  function calcMeanRates(rate, rateIsNull) result(meanRateStr)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: rate(:)
    logical,      intent(in) :: rateIsNull(size(rate))
    character(30)            :: meanRateStr

    real(single) :: meanRate
    character(30) :: rawMeanRateStr
    integer :: LengthOfStr

    !-------------------------------------------------------------------
    ! Start

    if (all(rateIsNull)) then
      meanRateStr = 'NULL'
    else
      meanRate = sum(rate, (.not.rateIsNull)) / real(count(.not.rateIsNull))
      if (meanRate < 0.0) then
        call string_of_real(abs(meanRate), rawMeanRateStr, LengthOfStr)
        meanRateStr = '-'//trim(rawMeanRateStr)
      else
        call string_of_real(meanRate, rawMeanRateStr, LengthOfStr)
        meanRateStr = ' '//trim(rawMeanRateStr)
      end if
    end if

  end function calcMeanRates

!-------------------------------------------------------------------------------------
  subroutine readgtis(GTI_File, gtis) 
  !-------------------------------------------------------------------------------------

! This subroutine reads the GTI file 

    character(len=*),intent(in) :: GTI_File
    type(IntervalT),pointer :: gtis(:)

! Local variables are Set, Table, Column and Pointer needed by DAL to recover
! FITS binary table. Moreover Time_Zero contains the start exposure time in MJD

    type( DataSetT ) :: Gti_Set
    type( TableT ) :: Gti_Table
    type( ColumnT ) :: Col_Start, Col_Stop
    real(kind=double),dimension(:),pointer :: Array_Start, Array_Stop
!    real(kind=single),dimension(:),pointer :: Gti_vals, Gti_vals_temp
!    real(kind=double),dimension(:),pointer :: Gti_times_temp, Gti_times
!    real(kind=double),dimension(:),pointer :: Time_Vect
!    real(kind=single) :: T_Start, T_Stop, time_exp, gtibinsize
    integer :: numofgtis
!, icnt, jcnt, Nbin_gtis, newbinfactor, kcnt

    Gti_Set = dataSet(GTI_File,READ)
    Gti_Table = table(Gti_Set,BlockNumber(Gti_Set,'STDGTI')) 

    Col_Start = Column(Gti_Table, ColumnNumber(Gti_Table,'START'),Read)
    Col_Stop = Column(Gti_Table, ColumnNumber(Gti_Table,'STOP'),Read)

   if ( .not. HasAttribute(Gti_Table,"TSTART") ) then
      call warning('noTstartKeyword','Keyword TSTART missing in '//GTI_File)
!    else
!      T_Start = real32Attribute(Gti_Table,'TSTART')
    endif

   if ( .not. HasAttribute(Gti_Table,"TSTOP") ) then
      call warning('noTstopKeyword','Keyword TSTOP missing in '//GTI_File)
!    else
!      T_Stop = real32Attribute(Gti_Table,'TSTOP')
    endif

!    binned = ' '

    Array_Start => real64Data(Col_Start)
    Array_Stop => real64Data(Col_Stop)

    numofgtis = size(Array_Start)
    allocate(gtis(numofgtis))
    gtis%lower = Array_Start
    gtis%upper = Array_Stop


    call release(Col_Start)
    call release(Col_Stop)
    call release(Gti_Table)
    call release(Gti_Set)

  end subroutine readgtis

!-------------------------------------------------------------------------------------
  subroutine readflare(Flare_File, Flare_Rate, Flare_Time) 
  !-------------------------------------------------------------------------------------

! This subroutine reads the flare file

    character(len=*),intent(in) :: Flare_File
    real(kind=single),dimension(:),pointer :: Flare_Rate
!, Flare_Time
    real(double),dimension(:),pointer :: Flare_Time
!    character(len=20),intent(out) :: binned
!    integer,intent(in) :: Nbin

! Local variables: Set, Table, Column and Pointer needed by DAL to recover FITS
! binary table. Moreover Time_Zero contains the start exposure time in MJD

    type( DataSetT ) :: Flare_Set
    type( TableT ) :: Flare_Table
    type( ColumnT ) :: Flare_Rate_col, Flare_Time_col
    real(kind=double),dimension(:),pointer :: Array_FTime
!, Flare_Time_temp
    real(kind=single),dimension(:),pointer :: Array_FRate
!    real(kind=double),dimension(:),pointer :: Gti_times, Time_Vect
!    real(kind=single),dimension(:),pointer :: Flare_Rate_temp, Flare_vals
!    real(kind=double),dimension(:),pointer :: single_FTime
!    real(kind=single) :: T_Start
    integer :: nflarebins
!, icnt, jcnt, Nbin_gtis, newbinfactor
    real(double) :: centreTime1stBin

    Flare_Set = dataSet(Flare_File,READ)
    Flare_Table = table(Flare_Set,BlockNumber(Flare_Set,'RATE')) 

    Flare_Rate_col = Column(Flare_Table, ColumnNumber(Flare_Table,'RATE'),Read)
    Flare_Time_col = Column(Flare_Table, ColumnNumber(Flare_Table,'TIME'),Read)


   if ( .not. HasAttribute(Flare_Table,"TSTART") ) then
      call warning('noTstartKeyword','Keyword TSTART missing in '//Flare_File)
!    else
!      T_Start = real32Attribute(Flare_Table,'TSTART')
    endif

    Array_FRate => real32Data(Flare_Rate_col)
    Array_FTime => real64Data(Flare_Time_col)
    nflarebins = size(Array_FRate)
    allocate(Flare_Rate(nflarebins),Flare_Time(nflarebins))
    Flare_Rate = Array_FRate
    Flare_Time = Array_FTime

    centreTime1stBin = calcCentreTime1stBin(Flare_Table)
    Flare_Time = Flare_Time + centreTime1stBin

    call release(Flare_Rate_col)
    call release(Flare_Time_col)
    call release(Flare_Table)
    call release(Flare_Set)

  end subroutine readflare

  !---------------------------------------------------------------------
  subroutine rebin_ims(rawData, numInBinsPerOutBin, plotData)
  !---------------------------------------------------------------------

    implicit none

    type(localDataType), intent(in)    :: rawData(:)
    integer,             intent(inout) :: numInBinsPerOutBin
    type(localDataType), pointer       :: plotData(:)

    integer ::&
      startInputBin, binOf1stNonNullBkgRate, numInputBins, numOutputBins,&
      inputBinStart, outputBinNum, i, numGoodInBins, lastInputBin,&
      oldNumInBinsPerOutBin
    real(double) :: timeSum
    real(single) :: fracSum, rateSum, errSquareSum, err

    !-------------------------------------------------------------------
    ! Start

    if (booleanParameter("forcestart")) then
      startInputBin = 1
    else ! start plot at first non-null bin:
      startInputBin = smallestTrueIndex(.not.rawData%net%isNull)
      ! smallestTrueIndex is in temp_ssclib.
      !
      ! We have already checked that not all the net rates were null,
      ! so this should work.

      if (.not.all(rawData%bkg%isNull)) then
        binOf1stNonNullBkgRate = smallestTrueIndex(.not.rawData%bkg%isNull)
        ! smallestTrueIndex is in temp_ssclib.
        startInputBin = min(startInputBin, binOf1stNonNullBkgRate)
      end if
    end if

    numInputBins = 1 + size(rawData) - startInputBin

    if (numInputBins > maxNumOutputBins * numInBinsPerOutBin) then
      oldNumInBinsPerOutBin = numInBinsPerOutBin
      numInBinsPerOutBin = ceiling(numInputBins / real(maxNumOutputBins))
      write(errStr, '(a, i3, a, i3, a)') "You asked for a rebinning &
        &ratio (= number of time series values averaged per plotted &
        &point) of ", oldNumInBinsPerOutBin, ". However this would &
        &result in too many points on the plot. elcplot has altered &
        &the bin ratio to ", numInBinsPerOutBin, "."
      call warning('rebinningSchemeAltered')
    end if
    numOutputBins = ceiling(numInputBins / real(numInBinsPerOutBin))

    allocate(plotData(numOutputBins))
!*** remember to deallocate in the calling program after you've finished with it!

    lastInputBin = size(rawData)
    plotData%net%isNull = .false. ! default
    plotData%bkg%isNull = .false. ! default
    inputBinStart = startInputBin - 1
    do outputBinNum = 1, numOutputBins
      if (inputBinStart + numInBinsPerOutBin > lastInputBin) then
        ! Don't average the time for the last bin, just extrapolate:
        plotData(outputBinNum)%time = plotData(outputBinNum-1)%time&
          + (plotData(outputBinNum-1)%time - plotData(1)%time)&
          / real(outputBinNum - 2, double)
      else
        ! Calc average time of the input bins:
        timeSum = 0.0d0
        do i = 1, numInBinsPerOutBin
          timeSum = timeSum + rawData(inputBinStart + i)%time
        end do
        plotData(outputBinNum)%time = timeSum&
          / real(numInBinsPerOutBin, double)
      end if

      ! Average the non-null src rates and errors (and fracexp):
      fracSum      = 0.0
      rateSum      = 0.0
      errSquareSum = 0.0
      numGoodInBins = 0

      do i = 1, numInBinsPerOutBin
      if (inputBinStart + i > lastInputBin) cycle
      if (rawData(inputBinStart + i)%net%isNull) cycle

        numGoodInBins = numGoodInBins + 1
        fracSum = fracSum + rawData(inputBinStart + i)%fracexp
        rateSum = rateSum + rawData(inputBinStart + i)%net%rate
        err     =           rawData(inputBinStart + i)%net%err
        errSquareSum = errSquareSum + err * err
      end do

      if (numGoodInBins > 0) then
        plotData(outputBinNum)%net%rate = rateSum&
          / real(numGoodInBins, single)
        plotData(outputBinNum)%net%err  = sqrt(errSquareSum)&
          / real(numGoodInBins, single)
      else
        plotData(outputBinNum)%net%isNull = .true.
      end if

      plotData(outputBinNum)%fracexp  = fracSum&
        / real(numInBinsPerOutBin, single)

      ! Average the non-null bkg rates and errors:
      rateSum      = 0.0
      errSquareSum = 0.0
      numGoodInBins = 0

      do i = 1, numInBinsPerOutBin
      if (inputBinStart + i > lastInputBin) cycle
      if (rawData(inputBinStart + i)%bkg%isNull) cycle

        numGoodInBins = numGoodInBins + 1
        rateSum = rateSum + rawData(inputBinStart + i)%bkg%rate
        err     =           rawData(inputBinStart + i)%bkg%err
        errSquareSum = errSquareSum + err * err
      end do

      if (numGoodInBins > 0) then
        plotData(outputBinNum)%bkg%rate = rateSum&
          / real(numGoodInBins, single)
        plotData(outputBinNum)%bkg%err  = sqrt(errSquareSum)&
          / real(numGoodInBins, single)
      else
        plotData(outputBinNum)%bkg%isNull = .true.
      end if

      inputBinStart = inputBinStart + numInBinsPerOutBin
    end do ! outputBinNum

  end subroutine rebin_ims

  !---------------------------------------------------------------------
  subroutine calcPgplotBoxWorldEdges(plotData, Flare_Rate&
  , timeAtLHWindowEdge&
  , timeAtRHWindowEdge, netRateAtWindowBot, netRateAtWindowTop&
  , bkgRateAtWindowBot, bkgRateAtWindowTop, fracExpAtWindowBot&
  , fracExpAtWindowTop, addnAtWindowBot, addnAtWindowTop)
  !---------------------------------------------------------------------

    implicit none

    type(localDataType), intent(in)  :: plotData(:)
    real(single),        intent(in)  :: Flare_Rate(:)
    real(double),        intent(out) :: timeAtLHWindowEdge,&
                                        timeAtRHWindowEdge
    real,                intent(out) :: netRateAtWindowBot,&
                                        netRateAtWindowTop,&
                                        bkgRateAtWindowBot,&
                                        bkgRateAtWindowTop,&
                                        fracExpAtWindowBot,&
                                        fracExpAtWindowTop,&
                                        addnAtWindowBot,&
                                        addnAtWindowTop

    real(double) :: timePadSize
    real(single) :: maxFracExp, fracExpPadSize, addnPadSize, maxFlareY
    integer :: numOutputBins, firstNonNullBin, i

    !-------------------------------------------------------------------
    ! Start

    numOutputBins = size(plotData)

    timePadSize = (plotData(numOutputBins)%time - plotData(1)%time)&
      * timePadFrac / (1.0d0 - 2.0d0 * timePadFrac)
    timeAtLHWindowEdge = plotData(1)%time             - timePadSize
    timeAtRHWindowEdge = plotData(numOutputBins)%time + timePadSize

    if (all(plotData%net%isNull))&
call error('allNetRatesNull',"After rebinning, all the net rate values &
&were found to be nulls")
    firstNonNullBin = smallestTrueIndex(.not.plotData%net%isNull)
    ! in temp_ssclib

    netRateAtWindowTop = plotData(firstNonNullBin)%net%rate&
      + plotData(firstNonNullBin)%net%err
    netRateAtWindowBot = plotData(firstNonNullBin)%net%rate&
      - plotData(firstNonNullBin)%net%err

    do i = firstNonNullBin + 1, numOutputBins
    if (plotData(i)%net%isNull) cycle
      netRateAtWindowTop = max(netRateAtWindowTop, plotData(i)%net%rate&
        + plotData(i)%net%err)
      netRateAtWindowBot = min(netRateAtWindowBot, plotData(i)%net%rate&
        - plotData(i)%net%err)
    end do

    if (.not. booleanParameter("bkgdyscale")) then
      if (all(plotData%bkg%isNull))&
call error('allBkgRatesNull',"After rebinning, all the background rate &
&values were found to be nulls")
      firstNonNullBin = smallestTrueIndex(.not.plotData%bkg%isNull)
      bkgRateAtWindowTop = plotData(firstNonNullBin)%bkg%rate&
        + plotData(firstNonNullBin)%bkg%err
      bkgRateAtWindowBot = plotData(firstNonNullBin)%bkg%rate&
        - plotData(firstNonNullBin)%bkg%err

      do i = firstNonNullBin + 1, numOutputBins
      if (plotData(i)%bkg%isNull) cycle
        bkgRateAtWindowTop = max(bkgRateAtWindowTop, plotData(i)%bkg%rate& 
          + plotData(i)%bkg%err)
        bkgRateAtWindowBot = min(bkgRateAtWindowBot, plotData(i)%bkg%rate&
          - plotData(i)%bkg%err)
      end do
    else ! make the y scale of bkg the same as for net.
      bkgRateAtWindowTop = netRateAtWindowTop
      bkgRateAtWindowBot = netRateAtWindowBot
    end if

    maxFracExp = maxval(plotData%fracexp)
    fracExpPadSize = maxFracExp * yPadFrac3rdPlot&
      / (1.0 - 2.0 * yPadFrac3rdPlot)
    fracExpAtWindowBot = 0.0
    fracExpAtWindowTop = maxFracExp + fracExpPadSize

    if (booleanParameter("usegtiset")) then
      addnPadSize = 1.0 * yPadFrac3rdPlot / (1.0 - 2.0 * yPadFrac3rdPlot)
      addnAtWindowBot = 0.0 - epsilon(1.0)
!      addnAtWindowTop = 1.0 + addnPadSize
      addnAtWindowTop = 2.0 
    else if (booleanParameter("useflareset")) then
      maxFlareY = maxval(Flare_Rate)
      addnPadSize = maxFlareY * yPadFrac3rdPlot&
        / (1.0 - 2.0 * yPadFrac3rdPlot)
      addnAtWindowBot = 0.0 - epsilon(1.0)
      addnAtWindowTop = maxFlareY + addnPadSize
    end if

  end subroutine calcPgplotBoxWorldEdges

  !---------------------------------------------------------------------
  subroutine calcTimeOffset(mjdRefDays, time1stBin, DATE_OBS, mjdOffsetDays&
  , offsetSec)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: mjdRefDays, time1stBin
    character(*), intent(in)  :: DATE_OBS
    real(double), intent(out) :: mjdOffsetDays, offsetSec

    real(double) :: offsetDays, mjdDateObsDays, mjdOffsetSec
    integer :: year, month, day, hour, min, sec, status
    real :: mjdOfDateObsFracDays
    character(len=4) :: offset_units

    !-------------------------------------------------------------------
    ! Start

    if (booleanParameter("offset")) then
      select case(stringParameter('offsetstyle'))
      case('auto')
        offsetSec = time1stBin
        mjdOffsetDays = mjdRefDays + offsetSec / 24.0d0 / 3600.0d0

      case('user')
        offset_units = StringParameter("ounits")

        select case(offset_units)
        case('secs')
          mjdOffsetSec = RealParameter("tscaleoffset")
          mjdOffsetDays = mjdOffsetSec / 24.0d0 / 3600.0d0
          offsetSec = mjdOffsetSec - mjdRefDays * 24.0d0 * 3600.0d0
        case('days')
          mjdOffsetDays = RealParameter("tscaleoffset")
          if (abs(mjdOffsetDays) >= 1.0e5)&
            call warning('checkOffsetUnit',"You have asked for a &
            &--tscaleoffset value of more than 10000, but the &
            &--ounits value is 'days'. Are you sure this is right?")

          offsetSec = (mjdOffsetDays - mjdRefDays) * 24.0d0 * 3600.0d0
        case default
call error('badOffsetUnit',"The value of --ounits was not recognized.")
        end select

      case('dateobs')
        if (len_trim(DATE_OBS) < 19) then
          write(errstr,*) "Bad format DATE-OBS keyword ", trim(DATE_OBS),  &
!                       " in ", trim(qualifiedName(Table1)),  &
                       ". Should be yyyy-mm-ddThh:mm:ss"
!          call warning("getXmmDate13")
call error("getXmmDate13")
!          return
        endif
        if (DATE_OBS( 5: 5) /= "-" .or. DATE_OBS( 8: 8) /= "-" .or.  &
          DATE_OBS(11:11) /= "T" .or.  &
          DATE_OBS(14:14) /= ":" .or. DATE_OBS(17:17) /= ":") then
          write(errstr,*) "Bad format DATE-OBS keyword ", trim(DATE_OBS),  &
!                       " in ", trim(qualifiedName(Table1)),  &
                       ". Should be yyyy-mm-ddThh:mm:ss"
!          call warning("getXmmDate13")
call error("getXmmDate13")
!          return
        endif
        read(DATE_OBS,100) year, month, day, hour, min, sec
100     format(I4,1x,I2,1x,I2,1x,I2,1x,I2,1x,I2)

        call sla_cldj(year, month, day, mjdDateObsDays, status)
        if (status > 0) then
          write(errstr,*) "Error", status, " in sla_cldj for DATE-OBS = ",  &
          trim(DATE_OBS),&
! " in ", trim(qualifiedName(Table1)),  &
          ". Should be yyyy-mm-ddThh:mm:ss"
!          call warning("getXmmDate13")
call error("getXmmDate13")
!          return
        endif
!        rsec = sec

        call sla_ctf2d(hour, min, real(sec), mjdOfDateObsFracDays, status)
        if (status > 0) then
          write(errstr,*) "Error", status, " in sla_ctf2d for DATE-OBS = ",  &
          trim(DATE_OBS),&
! " in ", trim(qualifiedName(Table1)),  &
          ". Should be yyyy-mm-ddThh:mm:ss"
!          call warning("getXmmDate13")
call error("getXmmDate13")
!          return
        endif

        mjdOffsetDays = mjdDateObsDays + mjdOfDateObsFracDays
        offsetDays = mjdOffsetDays - mjdRefDays
        offsetSec = offsetDays * 24.0d0 * 3600.0d0

      case default
call error('badOffsetStyle',"The value of --offsetstyle was not recognized.")
      end select

    else ! no offset:
      offsetSec = 0.0d0
      mjdOffsetDays = mjdRefDays
    end if

  end subroutine calcTimeOffset

!-------------------------------------------------------------------------------------
  subroutine create_labels(Input_File, plotbinsize, mjdOffsetDays, Labels)
  !-------------------------------------------------------------------------------------
  ! Generates labels to add on plot.
 
    character(*),intent(in) :: Input_File
    real,        intent(in) :: plotbinsize
    real(double),intent(in) :: mjdOffsetDays
    character(*),dimension(9),intent(out) :: Labels

    type( dataSetT ) :: Ts_set
    type( TableT ) :: Table1
    character(len=4) :: ra_hr_ch, ra_min_ch, dec_deg_ch, dec_sign
    character(len=4) :: dec_min_ch
    character(len=30) :: ra_sec_ch, dec_sec_ch
    character(len=12) :: OBS_ID, OBJECT, INSTRUME, EXP_ID, SRCNO
!, DATE_OBS, RA, DEC
    character(len=30) :: TIMEDEL, plotbinsizec, TSTART_MJDINT, TSTART_MJDDEC
    real(kind=double) :: TIME, ra_dec, dec_dec, TIME_OFFSETDEC, TIME_OFFSETINT
    real              :: ra_sec, dec_sec
    integer :: Length, ra_hr, ra_min, dec_deg, dec_min
    logical :: srcRaKwdFound, objRaKwdFound

    Ts_Set = dataSet(Input_File, READ)
    Table1 = table(Ts_Set, 'RATE') 

    if (HasAttribute(Ts_Set,"OBS_ID") ) then
         OBS_ID = StringAttribute(Ts_Set,'OBS_ID')
    else if (HasAttribute(Table1,"OBS_ID") ) then
         OBS_ID = StringAttribute(Table1,'OBS_ID')
    else
         OBS_ID = 'Indef'
    endif

    if (HasAttribute(Ts_Set,"OBJECT") ) then
         OBJECT = StringAttribute(Ts_Set,'OBJECT')
    else if (HasAttribute(Table1,"OBJECT") ) then
         OBJECT = StringAttribute(Table1,'OBJECT')
    else
         OBJECT = 'Indef'
    endif

    if (HasAttribute(Ts_Set,"INSTRUME") ) then
         INSTRUME = StringAttribute(Ts_Set,'INSTRUME')
    else if (HasAttribute(Table1,"INSTRUME") ) then
         INSTRUME = StringAttribute(Table1,'INSTRUME')
    else
         INSTRUME = 'Indef'
    endif

    if (trim(INSTRUME) == 'EPN') then
      INSTRUME = 'PN'
    else if (trim(INSTRUME) == 'EMOS1') then
      INSTRUME = 'M1'
    else if (trim(INSTRUME) == 'EMOS2') then
      INSTRUME = 'M2'
    else
         INSTRUME = INSTRUME
    endif

    if (HasAttribute(Ts_Set,"EXPIDSTR") ) then
         EXP_ID = StringAttribute(Ts_Set,'EXPIDSTR')
    else
         EXP_ID = 'Indef'
    endif

!    if (HasAttribute(Ts_Set,"SRCNO") ) then
    if (HasAttribute(Table1,"SRCNUM") ) then
!         SRCNO = StringAttribute(Ts_Set,'SRCNO')
         SRCNO = StringAttribute(Table1,'SRCNUM')
    else
         SRCNO = 'Indef'
    endif

    objRaKwdFound    = .false. ! default
    srcRaKwdFound = .false. ! default
    if (HasAttribute(Table1,"SRC_CRA") ) then
      srcRaKwdFound = .true.
      ra_dec = real64Attribute(Table1,'SRC_CRA')
    else if (HasAttribute(Table1,"RA_OBJ") ) then
      objRaKwdFound = .true.
      ra_dec = real64Attribute(Table1,'RA_OBJ')
    end if

    if (srcRaKwdFound .or. objRaKwdFound) then
      ra_hr = int(ra_dec/15.0)
      write(ra_hr_ch,'(I2)') ra_hr
      ra_min = int(((ra_dec/15.0)-ra_hr)*60.0)
      write(ra_min_ch,'(I2)') ra_min
      ra_sec = ((((ra_dec/15.0)-ra_hr)*60.0)-ra_min)*60.0
      if (ra_sec < 10.0) then
        call string_of_real(ra_sec, ra_sec_ch, Length)
        write(ra_sec_ch,'(A4)') ra_sec_ch
      else
        call string_of_real(ra_sec, ra_sec_ch, Length)
        write(ra_sec_ch,'(A5)') ra_sec_ch
     end if     
    else
      ra_hr_ch = '??' ;  ra_min_ch = '??'; ra_sec_ch = '??'
    end if

    if (srcRaKwdFound) then
      dec_dec = real64Attribute(Table1,'SRC_CDEC')
    else if (objRaKwdFound) then
      dec_dec = real64Attribute(Table1,'DEC_OBJ')
    end if

    if (srcRaKwdFound .or. objRaKwdFound) then
      if (dec_dec < 0.0 .and. dec_dec > -1.0) then 
        dec_sign = '-'
      else
        dec_sign = ' '
      end if
      dec_deg = int(dec_dec)
      write(dec_deg_ch,'(I3)') dec_deg
      dec_min= int(abs(((dec_dec)-dec_deg)*60.0))
      write(dec_min_ch,'(I2)') dec_min
      dec_sec = (abs(((dec_dec)-dec_deg)*60.0)-dec_min)*60.0
     if (dec_sec < 10.0) then
        call string_of_real(dec_sec, dec_sec_ch, Length)
        write(dec_sec_ch,'(A3)') dec_sec_ch
      else
        call string_of_real(dec_sec, dec_sec_ch, Length)
        write(dec_sec_ch,'(A4)') dec_sec_ch
      end if      
    else
      dec_deg_ch = '??' ;  dec_min_ch = '??'; dec_sec_ch = '??'
    end if

    if (HasAttribute(Ts_Set,"TIMEDEL") ) then
      TIME = real64Attribute(Ts_Set,'TIMEDEL')
      call string_of_real2(TIME, TIMEDEL, Length)
    else if (HasAttribute(Table1,"TIMEDEL") ) then
      TIME = real64Attribute(Table1,'TIMEDEL')
      call string_of_real2(TIME, TIMEDEL, Length)
    else
      TIMEDEL = 'Indef'
    endif

    call string_of_real2(real(plotbinsize, double), plotbinsizec, Length)

!    TIME_OFFSETINT=int(TIME_OFFSET)
    TIME_OFFSETINT=floor(abs(mjdOffsetDays))
!    TIME_OFFSETDEC=TIME_OFFSET-int(TIME_OFFSET)
    TIME_OFFSETDEC=abs(mjdOffsetDays)-TIME_OFFSETINT
    call string_of_real2(TIME_OFFSETINT, TSTART_MJDINT, Length)
    call string_of_real2(TIME_OFFSETDEC, TSTART_MJDDEC, Length)

    Labels(1) = 'Filename: '//trim(Input_File)
    Labels(2) = 'Obs. ID: '//trim(OBS_ID)//' ('//trim(OBJECT)// ')'
    Labels(3)= 'Exp ID: '//trim(INSTRUME)//trim(EXP_ID)
    Labels(4) ='Source no.: '//trim(SRCNO)
    Labels(5) = 'Source RA_corr: '&
      //trim(ra_hr_ch)//':'//trim(ra_min_ch)//':'//trim(ra_sec_ch)//' (J2000)'
    Labels(6) = 'Source Dec_corr: '&
      //trim(dec_sign)//trim(dec_deg_ch)//':'//trim(dec_min_ch)//':'//trim(dec_sec_ch)&
      //' (J2000)'
    Labels(7) = 'Time series bin size: '//trim(TIMEDEL)//' secs'
    Labels(8) = 'Plotted bin size: ' //trim(plotbinsizec)// ' secs'

    if (mjdOffsetDays < 0.0) then
      Labels(9) = 'Time series start (MJD): -'&
        //trim(TSTART_MJDINT)//TSTART_MJDDEC(2:10)
    else
      Labels(9) = 'Time series start (MJD): '&
        //trim(TSTART_MJDINT)//TSTART_MJDDEC(2:10)
    end if

    call release(Table1)
    call release(Ts_Set)

  end subroutine create_labels

!-------------------------------------------------------------------------------------
  subroutine open_plot(Device_Type, Output_File) 
  !-------------------------------------------------------------------------------------
  ! Check Device_Type and Output_File. Open plot if required.

    character(len=*),intent(inout) :: Device_Type , Output_File
!    integer :: pgbeg
integer :: pgopen
    character(len=2048) :: Device_Selected
    logical :: File_Exists

    if (Output_File == '')  Output_File='default.ps'
!    if (Device_Type(1:3) /= '/XW') then
if (Device_Type(1:3) /= '/XW' .and. Device_Type(1:3) /= '/XS') then
      inquire(file=Output_File,exist=File_Exists)   
      if (File_Exists .and. .not. clobber()) call error('noClobber', &
      'Existing file '//trim(Output_File)//' cannot be overwritten (SAS_CLOBBER=0)')
      if (File_Exists .and. clobber()) call warning('fileWillBeClobbered', &
      'Existing file '//trim(Output_File)//' will be overwritten')
      Device_Selected=trim(Output_File)//trim(Device_Type)
!      if ( pgbeg(0,Device_Selected,1,1)/=1 ) then
if ( pgopen(Device_Selected) <= 0 ) then
        call warning('badPgplotDevice', &
          'Invalid device and/or output file: '//trim(Device_Type)//' , '//trim(Output_File))
           Output_File='default.ps'
           Device_Type= '/PS'
           Device_Selected=trim(Output_File)//trim(Device_Type)
!           if ( pgbeg(0,Device_Selected,1,1)/=1 ) then
if ( pgopen(Device_Selected)<=0 ) then
             call error('badPgplotDevice', &
             'Invalid DEFAULT device and/or output file : '&
              //trim(Device_Type)//' , '//trim(Output_File))
           else
             call warning('newPgplotDevice', &
             'New device and output file: '//trim(Device_Type)//' , '//trim(Output_File))
            endif
         endif
    else
        Device_Selected=trim(Device_Type)
!        if ( pgbeg(0,Device_Selected,1,1)/=1 ) call error('ELCPLOT', &
if ( pgopen(Device_Selected)<=0 ) call error('badPgplotDevice', &
        'Invalid device : '//trim(Device_Type))
   endif

! Set up the page size - this is optimal for a portrait output
   call pgpap(0.0 , 1.618 )
!   call pgask(.true.)
!   call pgsch(0.9)

  end subroutine open_plot


  !---------------------------------------------------------------------
  subroutine plotHeaderInfo(Labels)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: Labels(9)

    integer :: i

    !-------------------------------------------------------------------
    ! Start

    info = getTaskInfo()

    call pgsch(0.9)
    call pgsvp(0.0,1.0,0.0,1.0)
    call pgsci(1)
    call pgmtxt("b", -1.0, 0.0, 0.0, trim(nameOf(info))//' v'&
                                               //trim(versionOf(info)))
       do i=1,6
!          call pgmtxt('T' , -(1. + (i-1)*1.2) , 0.5 , 0.5 , Labels(i))
          call pgmtxt('T' , -(1. + (i-1)*1.2) , 0.05 , 0.0 , Labels(i))
       end do
       do i=7,9
!          call pgmtxt('T' , -(1. + (i-4)*1.2) , 0.5 , 0.5 , Labels(i))
          call pgmtxt('T' , -(1. + (i-4)*1.2) , 0.55 , 0.0 , Labels(i))
       end do


  end subroutine plotHeaderInfo

  !---------------------------------------------------------------------
  subroutine make_one_plot(Window, timeAtLHWindowEdge, timeAtRHWindowEdge&
  , trueTimeAtLHWindowEdge, trueTimeAtRHWindowEdge&
  , yAtWindowBot, yWindowTop, boxXStr, boxYStr, xLabel, yLabel, Title)
  !---------------------------------------------------------------------

    implicit none

    real,         intent(in) :: Window(4),&
                                timeAtLHWindowEdge,&
                                timeAtRHWindowEdge,&
                                yAtWindowBot,&
                                yWindowTop
    real(double), intent(in) :: trueTimeAtLHWindowEdge,&
                                trueTimeAtRHWindowEdge
    character(*), intent(in) :: boxXStr,&
                                boxYStr,&
                                xLabel,&
                                yLabel,&
                                Title

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '  start of make_one_plot'

    call pgsvp(Window(1),Window(2),Window(3),Window(4))
    call pgswin(real(trueTimeAtLHWindowEdge), real(trueTimeAtRHWindowEdge)&
      , yAtWindowBot, yWindowTop)

    call pgsci(1)
    call pgslw(1)
    call pgsch(0.7)
    call pgbox(trim(boxXStr), 0.0, 0&
      , trim(boxYStr), 0.0, 0)
    call pgsch(0.9)
    call pglab(xLabel, yLabel, Title)

    call pgswin(timeAtLHWindowEdge, timeAtRHWindowEdge&
      , yAtWindowBot, yWindowTop)

  end subroutine make_one_plot

!  !---------------------------------------------------------------------
!  subroutine make_one_plot_new(Window, timeAtLHWindowEdge, timeAtRHWindowEdge&
!  , trueTimeAtLHWindowEdge, trueTimeAtRHWindowEdge&
!  , yAtWindowBot, yWindowTop, boxXStr, boxYStr, xLabel, yLabel, Title)
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    real,         intent(in) :: Window(4),&
!                                timeAtLHWindowEdge,&
!                                timeAtRHWindowEdge,&
!                                yAtWindowBot,&
!                                yWindowTop
!    real(double), intent(in) :: trueTimeAtLHWindowEdge,&
!                                trueTimeAtRHWindowEdge
!    character(*), intent(in) :: boxXStr,&
!                                boxYStr,&
!                                xLabel,&
!                                yLabel,&
!                                Title
!
!    real :: majorTickInterval
!    integer :: numMinorIntervals
!
!    !-------------------------------------------------------------------
!    ! Start
!
!!print *, '  make_one_plot start'
!
!    call pgsch(0.9)
!    call pgsvp(Window(1),Window(2),Window(3),Window(4))
!    call pgswin(timeAtLHWindowEdge, timeAtRHWindowEdge&
!      , yAtWindowBot, yWindowTop)
!
!call calcTicks(trueTimeAtLHWindowEdge, trueTimeAtRHWindowEdge&
!  , 9, majorTickInterval, numMinorIntervals)
!
!    call pgsci(1)
!    call pgslw(1)
!    call pgbox(trim(boxXStr), majorTickInterval, numMinorIntervals&
!      , trim(boxYStr), 0.0, 0)
!    call pglab(xLabel, yLabel, Title)
!
!call printNumTimeLabels(timeAtLHWindowEdge, timeAtRHWindowEdge&
!  , trueTimeAtLHWindowEdge, trueTimeAtRHWindowEdge, majorTickInterval)
!
!!print *, '  make_one_plot end'
!
!  end subroutine make_one_plot_new
!
!  !---------------------------------------------------------------------
!  subroutine calcTicks(tLo, tHi&
!  , maxNumMajorIntervals, majorTickInterval, numMinorIntervals)
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    real(double), intent(in)  :: tLo, tHi
!    integer,      intent(in)  :: maxNumMajorIntervals
!    real,         intent(out) :: majorTickInterval
!    integer,      intent(out) :: numMinorIntervals
!
!    real(double) :: logMinMajorTickInterval, logResidual
!    integer :: logInteger
!
!    !-------------------------------------------------------------------
!    ! Start
!
!    logMinMajorTickInterval = log10((tHi - tLo)&
!      / real(maxNumMajorIntervals, double))
!
!    logInteger = floor(logMinMajorTickInterval)
!    logResidual = logMinMajorTickInterval - logInteger
!    if (logResidual <= 0.0) then ! can't in theory be less, but it's nice to
!    ! cover all the possibilities.
!      majorTickInterval =  1.0 * 10.0**(real(logInteger))
!      numMinorIntervals = 2
!    else if (logResidual <= lnTwo/lnTen) then   !log10(2.0d0)) then
!      majorTickInterval =  2.0 * 10.0**(real(logInteger))
!      numMinorIntervals = 4
!    else if (logResidual <= lnFive/lnTen) then  !log10(5.0d0)) then
!      majorTickInterval =  5.0 * 10.0**(real(logInteger))
!      numMinorIntervals = 5
!    else
!      majorTickInterval = 10.0 * 10.0**(real(logInteger))
!      numMinorIntervals = 10
!    end if
!
!  end subroutine calcTicks
!
!  !---------------------------------------------------------------------
!  subroutine printNumTimeLabels(timeAtLHWindowEdge, timeAtRHWindowEdge&
!  , trueTimeAtLHWindowEdge, trueTimeAtRHWindowEdge, majorTickInterval)
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    real,         intent(in) :: timeAtLHWindowEdge,&
!                                timeAtRHWindowEdge,&
!                                majorTickInterval
!    real(double), intent(in) :: trueTimeAtLHWindowEdge,&
!                                trueTimeAtRHWindowEdge
!
!    !-------------------------------------------------------------------
!    ! Start
!
!    numLabels = 
!
!  end subroutine printNumTimeLabels

!-------------------------------------------------------------------------------------
  subroutine string_of_real(Nb , String , Length)
  !-------------------------------------------------------------------------------------
  ! Converts real numbers into plottable character chain. Gives also its length.

    real(kind=single),intent(in) :: Nb
    character(len=30),intent(out) :: String
    integer,intent(out) :: Length
    integer:: puiss
    integer::MM

!    puiss = max( 0 , min( 6 , 6-int(log10(Nb)) ) )
    puiss = max( 0 , min( 8 , 8-int(log10(Nb)) ) )
    MM = nint(Nb*(10**puiss))
    call pgnumb(MM , -puiss , 1 , String , Length)

  end subroutine string_of_real

  !-------------------------------------------------------------------------------------
  subroutine string_of_real2(Nb , String , Length)
  !-------------------------------------------------------------------------------------
  ! Converts real numbers into plottable character chain. Gives also its length.

    real(kind=double),intent(in) :: Nb
    character(len=30),intent(out) :: String
    integer,intent(out) :: Length
    integer:: puiss
    integer::MM

    puiss = max( 0 , min( 8 , 8-int(log10(Nb)) ) )
    MM = idnint(Nb*real(10**puiss))
    call pgnumb(MM , -puiss , 1 , String , Length)

  end subroutine string_of_real2


!-------------------------------------------------------------------------------------
  function change_case(instring,dir) result(outstring)
  !-------------------------------------------------------------------------------------
  ! Switch from lower to upper case or vice-versa: From Jean Ballet
  ! instring : input string
  ! dir      : direction (1: to lower case; 2: to uppercase; 0: switch case)

    character(len=*), intent(in) :: instring
    integer, intent(in) :: dir
    character(len=len(instring)) :: outstring

  ! ASCII positions of lower case and upper case letters

    integer, parameter :: start_up  = iachar("A"), end_up  = iachar("Z"), &
                          start_low = iachar("a"), end_low = iachar("z"), &
                          offset = start_low - start_up

    integer :: length, icur, pos

    length = len(instring)
    outstring = instring
    do icur = 1, length
       pos = iachar(instring(icur:icur))
       if (pos >= start_up .and. pos <= end_up .and. dir < 2) then
          pos = pos + offset
       else if (pos >= start_low .and. pos <= end_low .and. dir /= 1) then
          pos = pos - offset
       endif
       outstring(icur:icur) = achar(pos)
    enddo

  end function change_case      

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module elcplot_mod
